#  -*- coding: utf-8 -*-
# *****************************************************************************
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Module authors:
#   Markus Zolliker <markus.zolliker@psi.ch>
#
# *****************************************************************************

"""driver for phytron motors"""

import time
from frappy.core import Done, Command, EnumType, FloatRange, IntRange, \
    HasIO, Parameter, Property, Drivable, PersistentMixin, PersistentParam, \
    StringIO, StringType, IDLE, BUSY, ERROR, Limit
from frappy.errors import CommunicationFailedError, HardwareError
from frappy.features import HasOffset


class PhytronIO(StringIO):
    end_of_line = '\x03'  # ETX
    timeout = 0.5
    identification = [('0IVR', 'MCC Minilog .*')]

    def communicate(self, command):
        ntry = 5
        warn = None
        for itry in range(ntry):
            try:
                _, _, reply = super().communicate('\x02' + command).partition('\x02')
                if reply[0] == '\x06':  # ACK
                    break
                raise CommunicationFailedError('missing ACK %r (cmd: %r)'
                                               % (reply, command))
            except Exception as e:
                if itry < ntry - 1:
                    warn = e
                else:
                    raise
        if warn:
            self.log.warning('needed %d retries after %r', itry, warn)
        return reply[1:]


class Motor(HasOffset, PersistentMixin, HasIO, Drivable):
    axis = Property('motor axis X or Y', StringType(), default='X')
    address = Property('address', IntRange(0, 15), default=0)
    circumference = Property('cirumference for rotations or zero for linear', FloatRange(0), default=360)

    encoder_mode = Parameter('how to treat the encoder', EnumType('encoder', NO=0, READ=1, CHECK=2),
                             default=1, readonly=False)
    value = PersistentParam('angle', FloatRange(unit='deg'))
    status = PersistentParam()
    target = Parameter('target angle', FloatRange(unit='deg'), readonly=False)
    speed = Parameter('', FloatRange(0, 20, unit='deg/s'), readonly=False)
    accel = Parameter('', FloatRange(2, 250, unit='deg/s/s'), readonly=False)
    encoder_tolerance = Parameter('', FloatRange(unit='deg'), readonly=False, default=0.01)
    sign = PersistentParam('', IntRange(-1,1), readonly=False, default=1)
    encoder = Parameter('encoder reading', FloatRange(unit='deg'))
    backlash = Parameter("""backlash compensation\n
                         offset for always approaching from the same side""",
                         FloatRange(unit='deg'), readonly=False, default=0)
    target_min = Limit()
    target_max = Limit()
    alive_time = PersistentParam('alive time for detecting restarts',
                                 FloatRange(), default=0)  # export=False

    ioClass = PhytronIO
    fast_poll = 0.1
    _backlash_pending = False
    _mismatch_count = 0
    _step_size = None  # degree / step
    _reset_needed = False

    def get(self, cmd):
        return self.communicate('%x%s%s' % (self.address, self.axis, cmd))

    def set(self, cmd, value):
        # make sure e format is not used, max 8 characters
        strvalue = '%.6g' % value
        if 'e' in strvalue:
            if abs(value) <= 1:  # very small number
                strvalue = '%.7f' % value
            elif abs(value) < 99999999:  # big number
                strvalue = '%.0f' % value
            else:
                raise ValueError('number (%g) must not have more than 8 digits' % value)
        self.communicate('%x%s%s%s' % (self.address, self.axis, cmd, strvalue))

    def set_get(self, cmd, value, query):
        self.set(cmd, value)
        return self.get(query)

    def read_alive_time(self):
        now = time.time()
        axisbit = 1 << int(self.axis == 'Y')
        active_axes = int(self.get('P37R'))  # adr 37 is a custom address with no internal meaning
        if not (axisbit & active_axes):  # power cycle detected and this axis not yet active
            self.set('P37S', axisbit | active_axes)  # activate axis
            if now < self.alive_time + 7 * 24 * 3600:  # the device was running within last week
                # inform the user about the loss of position by the need of doing reset_error
                self._reset_needed = True
            else:  # do reset silently
                self.reset_error()
        self.alive_time = now
        self.saveParameters()
        return now

    def read_value(self):
        prev_enc = self.encoder
        pos = float(self.get('P20R')) * self.sign
        if self.encoder_mode != 'NO':
            enc = self.read_encoder()
        else:
            enc = pos
        status = self.communicate('%xSE' % self.address)
        status = status[0:4] if self.axis == 'X' else status[4:8]
        self.log.debug('run %s enc %s end %s', status[1], status[2], status[3])
        status = self.get('=H')
        if status == 'N':  # not at target
            if self.encoder_mode == 'CHECK':
                e1, e2 = sorted((prev_enc, enc))
                if e1 - self.encoder_tolerance <= pos <= e2 + self.encoder_tolerance:
                    self.status = BUSY, 'driving'
                else:
                    self.log.error('encoder lag: %g not within %g..%g',
                                   pos, e1, e2)
                    self.get('S')  # stop
                    self.status = ERROR, 'encoder lag error'
                    self.value = pos
                    self.saveParameters()
                    self.setFastPoll(False)
            else:
                self.status = BUSY, 'driving'
        else:
            if self._backlash_pending:
                # drive to real target
                self.set('A', self.sign * self.target)
                self._backlash_pending = False
                return pos
            if (self.encoder_mode == 'CHECK' and
                    abs(enc - pos) > self.encoder_tolerance):
                if self._mismatch_count < 2:
                    self._mismatch_count += 1
                else:
                    self.log.error('encoder mismatch: abs(%g - %g) < %g',
                                   enc, pos, self.encoder_tolerance)
                    self.status = ERROR, 'encoder does not match pos'
            else:
                self._mismatch_count = 0
                self.status = IDLE, ''
                self.value = pos
                self.saveParameters()
            self.setFastPoll(False)
        return pos

    def read_encoder(self):
        if self.encoder_mode == 'NO':
            return self.value
        return float(self.get('P22R')) * self.sign

    def write_sign(self, value):
        self.sign = value
        self.saveParameters()
        return Done

    def get_step_size(self):
        self._step_size = float(self.get('P03R'))

    def read_speed(self):
        if self._step_size is None:
            # avoid repeatedly reading step size, as this is polled and will not change
            self.get_step_size()
        return float(self.get('P14R')) * self._step_size

    def write_speed(self, value):
        self.get_step_size()  # read step size anyway, it does not harm
        return float(self.set_get('P14S', round(value / self._step_size), 'P14R')) * self._step_size

    def read_accel(self):
        if not self._step_size:
            self.get_step_size()
        return float(self.get('P15R')) * self._step_size

    def write_accel(self, value):
        self.get_step_size()
        return float(self.set_get('P15S', round(value / self._step_size), 'P15R')) * self._step_size

    def check_target(self, value):
        self.checkLimits(value)
        self.checkLimits(value + self.backlash)

    def write_target(self, value):
        self.read_alive_time()
        if self._reset_needed:
            self.status = ERROR, 'reset needed after power up (probably position lost)'
            raise HardwareError(self.status[1])
        if self.status[0] == ERROR:
            raise HardwareError('need reset')
        self.status = BUSY, 'changed target'
        self.saveParameters()
        if self.backlash:
            # drive first to target + backlash
            # we do not optimize when already driving from the right side
            self._backlash_pending = True
            self.set('A', self.sign * (value + self.backlash))
        else:
            self.set('A', self.sign * value)
        self.setFastPoll(True, self.fast_poll)
        return value

    def stop(self):
        self.get('S')

    @Command
    def reset_error(self):
        """Reset error, set position to encoder"""
        self.read_value()
        if self.status[0] == ERROR or self._reset_needed:
            newenc = enc = self.encoder
            pos = self.value
            if abs(enc - pos) > self.encoder_tolerance or self.encoder_mode == 'NO':
                if self.circumference:
                    # bring encoder value either within or as close as possible to the given range
                    if enc < self.target_min:
                        mid = self.target_min + 0.5 * min(self.target_max - self.target_min, self.circumference)
                    elif enc > self.target_max:
                        mid = self.target_max - 0.5 * min(self.target_max - self.target_min, self.circumference)
                    else:
                        mid = enc
                    newenc += round((mid - enc) / self.circumference) * self.circumference
                if newenc != enc:
                    self.set('P22S', newenc * self.sign)
            if newenc != pos:
                self.set('P20S', newenc * self.sign)  # set pos to encoder
            self.read_value()
            self._reset_needed = False

# TODO:
#  '=E' electronics status
#  '=I+' / '=I-': limit switches
